from itertools import product
from collections import namedtuple
import multiprocessing, time, os
import numpy as np

from .printer import Printer
from .headers import faces_header, owner_header, neighbour_header, boundary_header

Patch = namedtuple('Patch', ['name', 'type', 'startFace', 'nFaces'])
PatchSpec = namedtuple('PatchSpec', ['name', 'type', 'top_patch'])

class Face():
    def __init__(self, vertex, owner, neighbour=None):
        self.vertex = vertex
        self.owner = owner
        self.neighbour = neighbour


class FaceList():
    def __init__(self, isin, pointlist, celllist, cylinders, verbose):
        self.isin = isin
        self.pointlist = pointlist
        self.celllist = celllist
        self.cylinders = cylinders
        self._print = Printer(verbose)

        self._facelist = []
        self.patches = []
        self._build_list()

    def export(self, polyMesh_path):
        self._print("Exporting faces, owner and neighbour")
        self._export_faces(polyMesh_path)
        self._print("Exporting boundary")
        self._export_boundaries(polyMesh_path)

    def _num_internal_faces(self):
        n = 0
        for f in self._facelist:
            if f.neighbour is not None:
                n += 1
        return n

    def _export_faces(self, polyMesh_path):
        n_faces = len(self._facelist)

        faces_filepath = os.path.join(polyMesh_path, 'faces')
        f_faces = open(faces_filepath, 'w')
        f_faces.write(faces_header + '\n')
        f_faces.write(str(n_faces) + '\n')
        f_faces.write('(\n')

        owner_filepath = os.path.join(polyMesh_path, 'owner')
        f_owner = open(owner_filepath, 'w')
        f_owner.write(owner_header + '\n')
        f_owner.write(str(n_faces) + '\n')
        f_owner.write('(\n')

        neigh_filepath = os.path.join(polyMesh_path, 'neighbour')
        f_neigh = open(neigh_filepath, 'w')
        f_neigh.write(neighbour_header + '\n')
        f_neigh.write(str(self._num_internal_faces()) + '\n')
        f_neigh.write('(\n')

        for face in self._facelist:
            f_faces.write('4' + str(face.vertex).replace(',', '') + '\n')
            f_owner.write(str(face.owner) + '\n')
            if face.neighbour is not None:
                f_neigh.write(str(face.neighbour) + '\n')

        f_faces.write(')\n')
        f_owner.write(')\n')
        f_neigh.write(')\n')

        f_faces.close()
        f_owner.close()
        f_neigh.close()

    def _export_boundaries(self, polyMesh_path):
        bound_filepath = os.path.join(polyMesh_path, 'boundary')
        with open(bound_filepath, 'w') as fw:
            fw.write(boundary_header + '\n')
            fw.write(str(len(self.patches)) + "\n")
            fw.write("(\n")
            for pid, patch in enumerate(self.patches):
                fw.write(patch.name + "\n")
                fw.write("\t{\n")
                fw.write("\t\ttype       " + patch.type + ";\n")
                fw.write("\t\tnFaces     " + str(patch.nFaces)+ ";\n")
                fw.write("\t\tstartFace  " + str(patch.startFace)+ ";\n")
                fw.write("\t}\n")
            fw.write(")\n")

    def _build_list(self):
        self._print("Generating list of internal faces")
        self._get_internal_faces()
        self._print("Generating list of boundary faces")
        self._get_boundary_faces()

    def _get_internal_faces(self):
        nx, ny, nz = self.isin.shape
        n_cells = self.isin.size
        all_faces = []
        for n, (i, j, k) in enumerate(product(range(nx), range(ny), range(nz))):
            if (n+1) % 1000 == 0:
                prog = n / n_cells * 100.
                self._print(f'Reached cell {n+1} of {n_cells} ({prog:.2f}%)')
            if self.isin[i, j, k]:
                cell_add = (i, j, k)
                all_faces.append(self.celllist.get_cell_face(cell_add, 'up'))
                all_faces.append(self.celllist.get_cell_face(cell_add, 'north'))
                all_faces.append(self.celllist.get_cell_face(cell_add, 'east'))

        self._print("Removing boundaries from internal faces")
        internal_faces = [face for face in all_faces if face.neighbour is not None]
        self._facelist.extend(internal_faces)

    def _get_boundary_faces(self):
        # bottom most boundary
        self._print("Generating bottom most boundary")
        self._get_boundary_horizontal(0, 'down')

        # intermediate boundaries
        n_cyls = len(self.cylinders)
        l0 = 0
        for cyl in range(n_cyls):
            self._print(f"Generating boundaries for cylinder {cyl+1} of {n_cyls}")
            l1 = l0 + self.cylinders[cyl].n_layers
            layers = list(range(l0, l1))
            self._get_boundary_vertical(layers)
            l0 = l1

            if cyl < n_cyls - 1:
                if self.cylinders[cyl].diameter > self.cylinders[cyl + 1].diameter:
                    self._get_boundary_horizontal(l1 - 1, 'up')
                else:
                    self._get_boundary_horizontal(l1, 'down')

        # top most boundary
        self._print("Generating top most boundary")
        self._get_boundary_horizontal(l1 - 1, 'up')

    def _get_boundary_horizontal(self, layer, direction):
        assert direction in ['up', 'down']
        startFace = len(self._facelist)
        nFaces = 0
        nx, ny, nz = self.isin.shape
        k = layer
        for i, j in product(range(nx), range(ny)):
            if self.isin[i, j, k]:
                cell_add = (i, j, k)
                # For each of the four directions, check if cell is at the edge
                # of the grid or if it has no neighbour
                if direction=='up' and (k == nz - 1 or not self.isin[i, j, k + 1]):
                    face = self.celllist.get_cell_face(cell_add, 'up')
                    self._facelist.append(face)
                    nFaces += 1
                if direction=='down' and (k == 0 or not self.isin[i, j, k - 1]):
                    face = self.celllist.get_cell_face(cell_add, 'down')
                    self._facelist.append(face)
                    nFaces += 1
        newPatch = Patch(name='patch_' + str(len(self.patches)), type='patch',
            startFace=startFace, nFaces=nFaces)
        self.patches.append(newPatch)

    def _get_boundary_vertical(self, layers):
        startFace = len(self._facelist)
        nFaces = 0
        nx, ny, _ = self.isin.shape
        k0 = layers[0]
        for i, j in product(range(nx), range(ny)):
            if self.isin[i, j, k0]:
                # For each of the four directions, check if cell is at the edge
                # of the grid or if it has no neighbour
                boundary_directions = []
                if j == ny - 1 or not self.isin[i, j + 1, k0]:
                    boundary_directions.append('north')
                if i == nx - 1 or not self.isin[i + 1, j, k0]:
                    boundary_directions.append('east')
                if j == 0 or not self.isin[i, j - 1, k0]:
                    boundary_directions.append('south')
                if i == 0 or not self.isin[i - 1, j, k0]:
                    boundary_directions.append('west')
                for bd in boundary_directions:
                    for k in layers:
                        cell_add = (i, j, k)
                        face = self.celllist.get_cell_face(cell_add, bd)
                        self._facelist.append(face)
                        nFaces += 1
        newPatch = Patch(name='patch_' + str(len(self.patches)), type='patch',
            startFace=startFace, nFaces=nFaces)
        self.patches.append(newPatch)



import os
from itertools import product
import numpy as np

from .headers import point_header

class PointList():
    def __init__(self, isin, vertex):
        self.isin = isin
        self.vertex = vertex
        self._pointlist = self._build_list()  # list of vertex addresses (not space locations)
        self._pointarray = self._build_array()

    def _build_list(self):
        points = set()
        inc = [0, 1]
        nx, ny, nz = self.isin.shape
        for i, j, k in product(range(nx), range(ny), range(nz)):
            if self.isin[i, j, k]:
                for inc_x, inc_y, inc_z in product(inc, inc, inc):
                    new_address = (
                        i + inc_x,
                        j + inc_y,
                        k + inc_z)
                    points.add(new_address)
        return list(points)

    def _build_array(self):
        points = np.ones(self.vertex.shape[:-1], dtype=np.int) * -1  #TODO: replace this by a sparse matrix
        for n, p in enumerate(self._pointlist):
            points[p] = n
        return points

    def __getitem__(self, key):
        return self._pointlist[key]

    def __len__(self):
        return len(self._pointlist)

    def index(self, i, j, k):
        return self._pointarray[(i, j, k)]

    def export(self, polyMesh_path):
        filepath = os.path.join(polyMesh_path, 'points')
        with open(filepath, 'w') as fw:
            fw.write(point_header)
            fw.write("\n")
            fw.write(str(len(self._pointlist)) + "\n")
            fw.write("(\n")
            for point_add in self._pointlist:
                location = self.vertex[point_add]
                fw.write(str(tuple(location)).replace(',', '') + "\n")
            fw.write(")\n")
            fw.write("\n")



import math
from itertools import product
import numpy as np


class Cylinder():

    cell_edge = None

    def __init__(self, diameter:int, height:float, n_layers:int):
        """ Specifies a cylinder to be used to create the final volume (a stack of cylinders).

        Parameters
        ----------
        diameter : int
            Diameter of cylinder specified in cell edeges. Basically it tells how many
             cells will fit inside the cylinder diameter. This number must be odd
             because the axis of the cylinder contains the center of the central cell
             of each layer.
        height : float
            Height of the cylinder in meters.
        n_layers : int
            Into how many layers will the cylinder be divided. All layers will have equal
             layer_height of size height/n_layers meters.
        """

        if type(diameter) is not int:
            raise TypeError('diameter must be an integer')
        if diameter % 2 != 1 and diameter > 0:
            raise ValueError('diameter must be a positive odd number')
        if Cylinder.cell_edge is None:
            raise AttributeError('A value for Cylinder.edge must be given')

        self.diameter = diameter
        self.radius = diameter * Cylinder.cell_edge / 2.
        self.height = height
        self.n_layers = n_layers
        self.vertical_spacing = np.linspace(0, height, n_layers + 1)

    def who_is_in(self, center_locations):
        #TODO: can be improved by symmetry
        ni, nj, _ = center_locations.shape
        isin = np.zeros((ni, ni), dtype=bool)
        for i, j in product(range(ni), range(ni)):
            [cx, cy] = center_locations[i, j]
            dist = (cx ** 2. + cy ** 2.) ** .5
            isin[i, j] = self.radius >= dist
        return isin

    @classmethod
    def conv_diam(cls, diam_in_meters:float):
        """Converts a diameter given in meters to a diameter given in a valid number of
         cell edges.

        Parameters
        ----------
        diam_in_meters : float
            Diameter given in meters

        Returns
        -------
        int
            Diameter given in cell edges. The number will be odd as required.
        """
        if cls.cell_edge is None:
            raise AttributeError('A value for Cylinder.edge must be given')

        float_edges = diam_in_meters / cls.cell_edge
        int_edges = int(round(float_edges))
        if math.modf(float_edges)[0] > .5:
            return int_edges - 1 if int_edges % 2 == 0 else int_edges
        else:
            return int_edges + 1 if int_edges % 2 == 0 else int_edges


from itertools import product
import numpy as np

from .face import Face

class CellList():
    def __init__(self, isin, pointlist):
        self.isin = isin
        self.pointlist = pointlist
        self._celllist = self._build_list()
        self._cellarray = self._build_array()

    def get_cell_face(self, cell_address, direction):
        i, j, k = cell_address
        cell_from = np.array(cell_address)
        cell_from_index = self.index(cell_address)
        # We first assume that cell_to has a larger index than cell_from
        if direction == "up":
            cell_to = cell_from + [0, 0, 1]
            vertex = (
                self.pointlist.index(i,     j,     k + 1),
                self.pointlist.index(i + 1, j,     k + 1),
                self.pointlist.index(i + 1, j + 1, k + 1),
                self.pointlist.index(i,     j + 1, k + 1))
        elif direction == "down":
            cell_to = cell_from + [0, 0, -1]
            vertex = (
                self.pointlist.index(i,     j,     k),
                self.pointlist.index(i,     j + 1, k),
                self.pointlist.index(i + 1, j + 1, k),
                self.pointlist.index(i + 1, j,     k))
        elif direction == "west":
            cell_to = cell_from + [-1, 0, 0]
            vertex = (
                self.pointlist.index(i, j,     k    ),
                self.pointlist.index(i, j,     k + 1),
                self.pointlist.index(i, j + 1, k + 1),
                self.pointlist.index(i, j + 1, k    ))
        elif direction == "east":
            cell_to = cell_from + [1, 0, 0]
            vertex = (
                self.pointlist.index(i + 1, j,     k    ),
                self.pointlist.index(i + 1, j + 1, k    ),
                self.pointlist.index(i + 1, j + 1, k + 1),
                self.pointlist.index(i + 1, j,     k + 1))
        elif direction == "north":
            cell_to = cell_from + [0, 1, 0]
            vertex = (
                self.pointlist.index(i,     j + 1, k    ),
                self.pointlist.index(i,     j + 1, k + 1),
                self.pointlist.index(i + 1, j + 1, k + 1),
                self.pointlist.index(i + 1, j + 1, k    ))
        elif direction == "south":
            cell_to = cell_from + [0, -1, 0]
            vertex = (
                self.pointlist.index(i,     j, k    ),
                self.pointlist.index(i + 1, j, k    ),
                self.pointlist.index(i + 1, j, k + 1),
                self.pointlist.index(i,     j, k + 1))
        else:
            raise ValueError('Invalid direction')

        cell_to = tuple(cell_to)
        if self._is_cell_address_out_of_bounds(cell_to) or not self.isin[cell_to]:
            # This means we the face is a boundary
            return Face(vertex, owner=cell_from_index)
        else:
            cell_to_index = self.index(cell_to)
            if cell_to_index > cell_from_index:
                return Face(vertex, owner=cell_from_index, neighbour=cell_to_index)
            else:  # By the way this program was constructed, we probably will never hit this else
                return Face(vertex[::-1], owner=cell_to_index, neighbour=cell_from_index)

    def _is_cell_address_out_of_bounds(self, address):
        for ind, add_coord in enumerate(address):
            if add_coord < 0 or add_coord >= self.isin.shape[ind]:
                return True
        return False

    def _build_list(self):
        nx, ny, nz = self.isin.shape
        celllist = []
        for i, j, k in product(range(nx), range(ny), range(nz)):
            if self.isin[i, j, k]:
                celllist.append((i, j, k))
        return celllist

    def _build_array(self):
        cells = np.ones_like(self.isin, dtype=np.int) * -1  #TODO: replace this by a sparse matrix
        for n, c in enumerate(self._celllist):
            cells[c] = n
        return cells

    def __getitem__(self, key):
        return self._celllist[key]

    def __len__(self):
        return len(self._celllist)

    def index(self, key):
        return self._cellarray[key]



from datetime import datetime

class Printer():
    def __init__(self, verbose):
        self.verbose = verbose

    def __call__(self, text):
        if self.verbose:
            now = datetime.now()
            print(now.strftime("[%H:%M:%S]"), text)



from itertools import product
from pathlib import Path
from typing import List, Tuple
import numpy as np

from .cylinder import Cylinder
from .point import PointList
from .cell import CellList
from .face import FaceList, Patch, PatchSpec
from .printer import Printer

class Stack():
    def __init__(self, cylinders:List[Cylinder], verbose=False):
        """Specifies a volume that is made of a stack of cylinders

        Parameters
        ----------
        cylinders : List[Cylinder]
            A list of cylinders to stack. cylinders[0] is the bottom most cylinder, and
             cylinders[-1] is the top most
        verbose : bool, optional
            Print runtimet messages, by default False
        """

        self.edge = Cylinder.cell_edge
        self.cylinders = cylinders
        self._print = Printer(verbose)

        self.max_diam = max([c.diameter for c in self.cylinders])

        self._print("Generating list of active cells")
        self.isin = self._who_is_in()
        self._print("Generating wireframe")
        self.vertex = self._build_vertex()
        self._print("Generating list of active points")
        self.pointlist = PointList(self.isin, self.vertex)
        self._print("Indexing active cells")
        self.celllist = CellList(self.isin, self.pointlist)
        self._print("Number of active cells " + str(len(self.celllist)) + " of " + str(self.isin.flatten().shape[0]))
        self._print("Generating list of faces")
        self.facelist = FaceList(self.isin, self.pointlist, self.celllist, self.cylinders, verbose)

    @property
    def n_patches(self):
        """
        Number of patches
        """
        return len(self.facelist.patches)

    def name_patches(self, patch_specs:List[PatchSpec]):
        """Group patches, give them names and assing their types.

        Parameters
        ----------
        patch_specs : List[PatchSpec]
            A PatchSpec is a tupple containing (name, type, last_patch).
             * name is the patch name and can be anything e.g. nozzle
             * type is any valid OpenFoam boundary type, e.g wall
             * last_patch is the index of last layer that will compose the grouped patch.
               The patches between patch_specs[n-1][2] and patch_specs[n][2] will be
               lumped together into a single patch.
        """

        new_patches = []
        old_patches = self.facelist.patches
        if len(old_patches) - 1 != patch_specs[-1].top_patch:
            raise ValueError(f"The top_patch of the last patch should be {len(old_patches) - 1}")
        for i in range(len(patch_specs)-1):
            if patch_specs[i+1].top_patch <= patch_specs[i].top_patch:
                raise ValueError("Top patches should be ordered in ascending order")

        base_patch = 0
        for pspec in patch_specs:
            patches_to_merge = old_patches[base_patch : pspec.top_patch+1]
            startFace = patches_to_merge[0].startFace
            nFaces = sum([p.nFaces for p in patches_to_merge])
            newPatch = Patch(name=pspec.name, type=pspec.type, startFace=startFace, nFaces=nFaces)
            new_patches.append(newPatch)
            base_patch = pspec.top_patch + 1

        self.facelist.patches = new_patches

    def export(self, filepath):
        Path(filepath).mkdir(parents=True, exist_ok=True)
        self._print("Exporting point list")
        self.pointlist.export(filepath)
        self._print("Exporting face list")
        self.facelist.export(filepath)
        self._print("Done exporting")

    def _who_is_in(self):
        h_max = (self.max_diam - 1) * self.edge / 2.
        h_min = -h_max
        horiz_spacing = np.linspace(h_min, h_max, self.max_diam)

        cx, cy = np.meshgrid(horiz_spacing, horiz_spacing)
        centers_2D = np.array([cx, cy])
        centers_2D = np.moveaxis(centers_2D, 0, -1)

        n_layers = sum([c.n_layers for c in self.cylinders])

        isin = np.zeros((self.max_diam, self.max_diam, n_layers), dtype=bool)
        k = 0
        for c in self.cylinders:
            c_isin = c.who_is_in(centers_2D)
            isin[:, :, k:k+c.n_layers] = c_isin[:, :, np.newaxis]
            k += c.n_layers

        return isin

    def _build_vertex(self):
        h_max = self.max_diam * self.edge / 2.
        h_min = -h_max
        horiz_spacing = np.linspace(h_min, h_max, self.max_diam + 1)

        vert_spacing = np.array([])
        height_shift = 0
        for c in self.cylinders:
            cyl_vert_spa = c.vertical_spacing[:-1] + height_shift
            vert_spacing = np.hstack((vert_spacing, cyl_vert_spa))
            height_shift += c.height
        vert_spacing = np.hstack((vert_spacing, height_shift))

        vx, vy, vz = np.meshgrid(horiz_spacing, horiz_spacing, vert_spacing, indexing='ij')
        vertex = np.array([vx, vy, vz])
        vertex = np.moveaxis(vertex, 0, -1)

        return vertex

